<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Feature;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;
use Org\Mxchange\CoreFramework\Factory\ObjectFactory;
use Org\Mxchange\CoreFramework\Loader\NoClassException;
use Org\Mxchange\CoreFramework\Object\BaseFrameworkSystem;
use Org\Mxchange\CoreFramework\Utils\String\StringUtils;

/**
 * The general feature management class. No instance is needed as this class
 * has only public methods that are static.
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2020 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class FrameworkFeature extends BaseFrameworkSystem {
	// Exception code constants
	const EXCEPTION_FEATURE_METHOD_NOT_CALLABLE = 0x400;

	/**
	 * "Cache" for enabled, available feature instances
	 *
	 * A typical available entry looks like this:
	 *
	 * array(
	 *     'is_enabled'   => true,
	 *     'is_available' => true,
	 *     'instance'     => SomeFeature Object
	 * )
	 *
	 * And a typical disabled entry looks like this:
	 *
	 * array(
	 *     'is_enabled'   => false,
	 *     'is_available' => false,
	 *     'instance'     => NULL
	 * )
	 */
	private static $enabledFeatures = [];

	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	protected function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Checks whether the given feature is enabled in configuration. The user
	 * shall be able to disable features, even when they *could* be available.
	 *
	 * @param	$featureName	Name of the feature to be checked
	 * @return	$isEnabled		Whether the given feature is enabled
	 */
	public static function isFeatureEnabled ($featureName) {
		// Is the cache set?
		if (!isset(self::$enabledFeatures[$featureName]['is_enabled'])) {
			// Generate config key
			$configKey = sprintf('enable_feature_%s', $featureName);

			// Check configuration
			self::$enabledFeatures[$featureName]['is_enabled'] = (FrameworkBootstrap::getConfigurationInstance()->getConfigEntry($configKey) === 'Y');
		} // END - if

		// Return "cached" status
		return self::$enabledFeatures[$featureName]['is_enabled'];
	}

	/**
	 * Checks whether the given feature is enabled and available. It may be
	 * enabled by the user, but is not available due to e.g. a missing PECL
	 * extension or whatever is needed to have this feature available. If you
	 * don't write a pre filters for checking PHP requirements, this is the
	 * method you want to use.
	 *
	 * @param	$featureName	Name of the feature to be checked on availability
	 * @return	$isAvailable	Whether the given feature is available
	 */
	public static function isFeatureAvailable ($featureName) {
		// Is the cache set?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FRAMEWORK-FEATURE: featureName=%s - CALLED!', $featureName));
		if (!isset(self::$enabledFeatures[$featureName]['is_available'])) {
			// Default is not available
			self::$enabledFeatures[$featureName]['is_available'] = false;
			self::$enabledFeatures[$featureName]['instance']     = NULL;

			// Is the feature enabled?
			if (!self::isFeatureEnabled($featureName)) {
				// Then it can't be available
				self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FRAMEWORK-FEATURE: Feature "%s"is not enabled.', $featureName));
				return false;
			} // END - if

			// Create config key (for feature class lookup)
			$configKey = sprintf('feature_%s_class', $featureName);

			// Now try to get the instance
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FRAMEWORK-FEATURE: configKey=%s', $configKey));
			try {
				// Try to get an instance
				self::$enabledFeatures[$featureName]['instance'] = ObjectFactory::createObjectByConfiguredName($configKey);

				// Now let the feature test itself's availability
				self::$enabledFeatures[$featureName]['is_available'] = self::$enabledFeatures[$featureName]['instance']->isFeatureAvailable();
			} catch (NoClassException $e) {
				// Feature class not found
				self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FRAMEWORK-FEATURE: Feature "%s"is not available due to missing feature class. Disabling feature ...', $featureName));
			}
		} // END - if

		// Return "cached" status
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FRAMEWORK-FEATURE: featureName=%s,isAvailable=%d - EXIT!', $featureName, intval(self::$enabledFeatures[$featureName]['is_available'])));
		return self::$enabledFeatures[$featureName]['is_available'];
	}

	/**
	 * Calls the feature's method and handles some arguments (if not given,
	 * NULL) to it. Any returned value is being forwarded to the caller, even
	 * when the doc-tag says 'void' as returned value.
	 *
	 * @param	$featureName	Name of the feature, it must be available at this point
	 * @param	$featureMethod	Method name of the feature's class
	 * @param	$args			Any arguments that should be handled over
	 * @return	$return			Anything the feature's method has returned
	 * @throws	FeatureMethodNotCallableException	If the requested method cannot be called
	 */
	public static function callFeature ($featureName, $featureMethod, array $args = NULL) {
		/*
		 * Please make sure that isFeatureAvailable() has been called and it has
		 * returned true before calling this method.
		 */
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FRAMEWORK-FEATURE: featureName=%s,featureMethod=%s,args[]=%s - CALLED!', $featureName, $featureMethod, gettype($args)));
		assert(self::isFeatureAvailable($featureName));

		// Array for call-back
		$callable = array(
			self::$enabledFeatures[$featureName]['instance'],
			sprintf('featureMethod%s', StringUtils::convertToClassName($featureMethod))
		);

		// So is the feature's method callable?
		if (!is_callable($callable)) {
			// Not callable method requested
			throw new FeatureMethodNotCallableException(array(self::$enabledFeatures[$featureName]['instance'], $featureMethod), self::EXCEPTION_FEATURE_METHOD_NOT_CALLABLE);
		} // END - if

		// Then call it
		$return = call_user_func_array($callable, $args);

		// Return any returned value
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FRAMEWORK-FEATURE: return[]=%s - EXIT!', gettype($return)));
		return $return;
	}

}
