<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Filesystem\File;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;
use Org\Mxchange\CoreFramework\Factory\ObjectFactory;
use Org\Mxchange\CoreFramework\Filesystem\Block;
use Org\Mxchange\CoreFramework\Filesystem\Block\CalculatableBlock;
use Org\Mxchange\CoreFramework\Filesystem\File\BaseAbstractFile;

// Import SPL stuff
use \BadMethodCallException;
use \InvalidArgumentException;
use \SplFileInfo;

/**
 * A general binary file class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2020 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
abstract class BaseBinaryFile extends BaseAbstractFile {
	/**
	 * Separator for header data
	 */
	const SEPARATOR_HEADER_DATA = 0x01;

	/**
	 * Separator header->entries
	 */
	const SEPARATOR_HEADER_ENTRIES = 0x02;

	/**
	 * Separator group->hash
	 */
	const SEPARATOR_GROUP_HASH = 0x03;

	/**
	 * Separator hash->value
	 */
	const SEPARATOR_HASH_VALUE = 0x04;

	/**
	 * Separator entry->entry
	 */
	const SEPARATOR_ENTRIES = 0x05;

	/**
	 * Separator type->position
	 */
	const SEPARATOR_TYPE_POSITION = 0x06;

	/**
	 * Length of count
	 */
	const LENGTH_COUNT = 20;

	/**
	 * Length of position
	 */
	const LENGTH_POSITION = 20;

	/**
	 * Length of group
	 */
	const LENGTH_GROUP = 10;

	/**
	 * Maximum length of entry type
	 */
	const LENGTH_TYPE = 20;

	//***** Array elements for 'gaps' array *****

	/**
	 * Start of gap
	 */
	const GAPS_INDEX_START = 'start';

	/**
	 * End of gap
	 */
	const GAPS_INDEX_END = 'end';

	/**
	 * Current seek position
	 */
	private $seekPosition = 0;

	/**
	 * Size of header
	 */
	private $headerSize = 0;

	/**
	 * File header
	 */
	private $header = [];

	/**
	 * Seek positions for gaps ("fragmentation")
	 */
	private $gaps = [];

	/**
	 * Seek positions for damaged entries (e.g. mismatching hash sum, ...)
	 */
	private $damagedEntries = [];

	/**
	 * Back-buffer
	 */
	private $backBuffer = '';

	/**
	 * Currently loaded block (will be returned by current())
	 */
	private $currentBlock = '';

	/**
	 * An instance of a Block class
	 */
	private $blockInstance = NULL;

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the class
	 * @return	void
	 */
	protected function __construct (string $className) {
		// Call parent constructor
		parent::__construct($className);

		// Init counters and gaps array
		$this->initCountersGapsArray();
	}

	/**
	 * Setter for Block instance
	 *
	 * @param	$blockInstance	An instance of an Block class
	 * @return	void
	 */
	protected final function setBlockInstance (Block $blockInstance) {
		$this->blockInstance = $blockInstance;
	}

	/**
	 * Getter for Block instance
	 *
	 * @return	$blockInstance	An instance of an Block class
	 */
	public final function getBlockInstance () {
		return $this->blockInstance;
	}

	/**
	 * Checks whether the abstracted file only contains gaps by counting all
	 * gaps' bytes together and compare it to total length.
	 *
	 * @return	$isGapsOnly		Whether the abstracted file only contains gaps
	 */
	private function isFileOnlyGaps () {
		// First/last gap found?
		/* Only for debugging
		if (isset($this->gaps[0])) {
			// Output first and last gap
			self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('[%s:%d:] this->gaps[0]=%s,this->gaps[%s]=%s', print_r($this->gaps[0], true), (count($this->gaps) - 1), print_r($this->gaps[count($this->gaps) - 1], true)));
		}
		*/

		// Now count every gap
		$gapsSize = 0;
		foreach ($this->gaps as $gap) {
			// Calculate size of found gap: end-start including both
			$gapsSize += ($gap[self::GAPS_INDEX_END] - $gap[self::GAPS_INDEX_START]);
		}

		// Debug output
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('[%s:%d:] gapsSize=%s,this->headerSize=%s', $gapsSize, $this->getHeaderSize()));

		// Total gap size + header size must be same as file size
		$isGapsOnly = (($this->getHeaderSize() + $gapsSize) == $this->getFileSize());

		// Return status
		return $isGapsOnly;
	}

	/**
	 * Initializes counter for valid entries, arrays for damaged entries and
	 * an array for gap seek positions. If you call this method on your own,
	 * please re-analyze the file structure. So you are better to call
	 * analyzeFile() instead of this method.
	 *
	 * @return	void
	 */
	public function initCountersGapsArray () {
		// Init counter and seek position
		$this->setCounter(0);
		$this->setSeekPosition(0);

		// Init arrays
		$this->gaps = [];
		$this->damagedEntries = [];
	}

	/**
	 * Getter for header size
	 *
	 * @return	$totalEntries	Size of file header
	 */
	public final function getHeaderSize () {
		// Get it
		return $this->headerSize;
	}

	/**
	 * Setter for header size
	 *
	 * @param	$headerSize		Size of file header
	 * @return	void
	 */
	public final function setHeaderSize (int $headerSize) {
		// Set it
		$this->headerSize = $headerSize;
	}

	/**
	 * Getter for header array
	 *
	 * @return	$totalEntries	Size of file header
	 */
	public final function getHeader () {
		// Get it
		return $this->header;
	}

	/**
	 * Setter for header
	 *
	 * @param	$header		Array for a file header
	 * @return	void
	 */
	public final function setHeader (array $header) {
		// Set it
		$this->header = $header;
	}

	/**
	 * Getter for seek position
	 *
	 * @return	$seekPosition	Current seek position (stored here in object)
	 */
	public final function getSeekPosition () {
		// Get it
		return $this->seekPosition;
	}

	/**
	 * Setter for seek position
	 *
	 * @param	$seekPosition	Current seek position (stored here in object)
	 * @return	void
	 */
	protected final function setSeekPosition (int $seekPosition) {
		// And set it
		$this->seekPosition = $seekPosition;
	}

	/**
	 * Updates seekPosition attribute from file to avoid to much access on file.
	 *
	 * @return	void
	 */
	public function updateSeekPosition () {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('[%s:%d:] CALLED!'));

		// Get key (= seek position)
		$seekPosition = $this->key();
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('[%s:%d:] Setting seekPosition=%s', $seekPosition));

		// And set it here
		$this->setSeekPosition($seekPosition);

		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('[%s:%d:] EXIT!'));
	}

	/**
	 * Seeks to beginning of file, updates seek position in this object and
	 * flushes the header.
	 *
	 * @return	void
	 */
	protected function rewindUpdateSeekPosition () {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('[%s:%d:] CALLED!'));

		// flushFileHeader must be callable
		assert(is_callable(array($this, 'flushFileHeader')));

		// Seek to beginning of file
		$this->rewind();

		// And update seek position ...
		$this->updateSeekPosition();

		// ... to write it back into the file
		$this->flushFileHeader();

		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('[%s:%d:] EXIT!'));
	}

	/**
	 * Seeks to old position
	 *
	 * @return	void
	 */
	protected function seekToOldPosition () {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('[%s:%d:] CALLED!'));

		// Seek to currently ("old") saved position
		$this->seek($this->getSeekPosition());

		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('[%s:%d:] EXIT!'));
	}

	/**
	 * Checks whether the block separator has been found
	 *
	 * @param	$str		String to look in
	 * @return	$isFound	Whether the block separator has been found
	 */
	public static function isBlockSeparatorFound ($str) {
		// Determine it
		$isFound = (strpos($str, chr(self::SEPARATOR_ENTRIES)) !== false);

		// Return result
		return $isFound;
	}

	/**
	 * Initializes the back-buffer by setting it to an empty string.
	 *
	 * @return	void
	 */
	private function initBackBuffer () {
		// Simply call the setter
		$this->setBackBuffer('');
	}

	/**
	 * Setter for backBuffer field
	 *
	 * @param	$backBuffer		Characters to "store" in back-buffer
	 * @return	void
	 */
	private function setBackBuffer (string $backBuffer) {
		// ... and set it
		$this->backBuffer = $backBuffer;
	}

	/**
	 * Getter for backBuffer field
	 *
	 * @return	$backBuffer		Characters "stored" in back-buffer
	 */
	private function getBackBuffer () {
		return $this->backBuffer;
	}

	/**
	 * Setter for currentBlock field
	 *
	 * @param	$currentBlock		Characters to set a currently loaded block
	 * @return	void
	 */
	private function setCurrentBlock (string $currentBlock) {
		// ... and set it
		$this->currentBlock = $currentBlock;
	}

	/**
	 * Gets currently read data
	 *
	 * @return	$current	Currently read data
	 */
	public function getCurrentBlock () {
		// Return it
		return $this->currentBlock;
	}

	/**
	 * Initializes this file class
	 *
	 * @param	$fileInfoInstance	An instance of a SplFileInfo class
	 * @return	void
	 */
	protected function initFile (SplFileInfo $fileInfoInstance) {
		// Get a file i/o pointer instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: fileInfoInstance[%s]=%s - CALLED!', get_class($fileInfoInstance), $fileInfoInstance));
		$pointerInstance = ObjectFactory::createObjectByConfiguredName('file_raw_input_output_class', array($fileInfoInstance));

		// ... and set it here
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Setting pointerInstance=%s ...', $pointerInstance->__toString()));
		$this->setPointerInstance($pointerInstance);

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Writes data at given position
	 *
	 * @param	$seekPosition	Seek position
	 * @param	$data			Data to be written
	 * @param	$flushHeader	Whether to flush the header (default: flush)
	 * @return	void
	 * @throws	InvalidArgumentException	If a parameter is invalid
	 */
	public function writeData (int $seekPosition, string $data, bool $flushHeader = true) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: seekPosition=%s,data()=%d,flushHeader=%d - CALLED!', $seekPosition, strlen($data), intval($flushHeader)));
		if ($seekPosition < 0) {
			// Invalid seek position
			throw new InvalidArgumentException(sprintf('seekPosition=%d is not valid', $seekPosition));
		} elseif (empty($data)) {
			// Empty data is invalid, too
			throw new InvalidArgumentException('Parameter "data" is empty');
		}

		// Write data at given position
		$this->getPointerInstance()->writeAtPosition($seekPosition, $data);

		// Increment counter
		$this->incrementCounter();

		// Update seek position
		$this->updateSeekPosition();

		// Flush the header?
		if ($flushHeader === true) {
			// Flush header
			$this->flushFileHeader();

			// Seek to old position
			$this->seekToOldPosition();
		}

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Marks the currently loaded block as empty (with length of the block)
	 *
	 * @param	$length		Length of the block
	 * @return	void
	 * @throws	InvalidArgumentException	If a parameter is invalid
	 */
	protected function markCurrentBlockAsEmpty (int $length) {
		// Validate parameter
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: length=%d - CALLED!', $length));
		if ($length < 1) {
			// Length cannot below one
			throw new InvalidArgumentException(sprintf('length=%d is not valid', $length));
		}

		// Get current seek position
		$currentPosition = $this->key();

		// Now add it as gap entry
		array_push($this->gaps, array(
			self::GAPS_INDEX_START  => ($currentPosition - $length),
			self::GAPS_INDEX_END    => $currentPosition,
		));

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Checks whether the file header is initialized
	 *
	 * @return	$isInitialized	Whether the file header is initialized
	 */
	public function isFileHeaderInitialized () {
		// Default is not initialized
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$isInitialized = false;

		// Is the file initialized?
		if ($this->isFileInitialized()) {
			// Some bytes has been written, so rewind to start of it.
			$rewindStatus = $this->rewind();

			// Is the rewind() call successfull?
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: rewindStatus=%d', $rewindStatus));
			if ($rewindStatus != 1) {
				// Something bad happened
				self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Could not rewind().');
			}

			// Read file header
			$this->readFileHeader();

			// The above method does already check the header
			$isInitialized = true;
		}

		// Return result
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: isInitialized=%d - EXIT!', intval($isInitialized)));
		return $isInitialized;
	}

	/**
	 * Checks whether the assigned file has been initialized
	 *
	 * @return	$isInitialized		Whether the file's size is zero
	 */
	public function isFileInitialized () {
		// Get it from iterator which holds the pointer instance. If false is returned
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$fileSize = $this->size();
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: fileSize=%s', $fileSize));

		/*
		 * The returned file size should not be false or NULL as this means
		 * that the pointer class does not work correctly.
		 */
		assert(is_int($fileSize));

		// Is more than 0 returned?
		$isInitialized = ($fileSize > 0);

		// Return result
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: isInitialized=%d - EXIT!', intval($isInitialized)));
		return $isInitialized;
	}

	/**
	 * Creates the assigned file
	 *
	 * @return	void
	 */
	public function createFileHeader () {
		// The file's header should not be initialized here
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		assert(!$this->isFileHeaderInitialized());

		// Simple flush file header which will create it.
		$this->flushFileHeader();

		// Rewind seek position (to beginning of file) and update/flush file header
		$this->rewindUpdateSeekPosition();

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Pre-allocates file (if enabled) with some space for later faster write access.
	 *
	 * @param	$type	Type of the file
	 * @return	void
	 * @throws	InvalidArgumentException	If a parameter is empty
	 */
	public function preAllocateFile (string $type) {
		// Is it enabled?
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: type=%s - CALLED!', $type));
		if (empty($type)) {
			// Empty type
			throw new InvalidArgumentException('Parameter "type" is empty');
		} elseif (FrameworkBootstrap::getConfigurationInstance()->getConfigEntry($type . '_pre_allocate_enabled') != 'Y') {
			// Don't continue here.
			self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Not pre-allocating file.'));
			return;
		}

		// Message to user
		self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: Pre-allocating file ...');

		// Calculate minimum length for one entry
		$minLengthEntry = $this->getBlockInstance()->calculateMinimumBlockLength();

		// Calulcate seek position
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: minLengthEntry=%s', $minLengthEntry));
		$seekPosition = $minLengthEntry * FrameworkBootstrap::getConfigurationInstance()->getConfigEntry($type . '_pre_allocate_count');

		// Now simply write a NUL there. This will pre-allocate the file.
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: seekPosition=%d', $seekPosition));
		$this->writeData($seekPosition, chr(0));

		// Rewind seek position (to beginning of file) and update/flush file header
		$this->rewindUpdateSeekPosition();

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Determines seek position
	 *
	 * @return	$seekPosition	Current seek position
	 */
	public function determineSeekPosition () {
		// Call pointer instance
		return $this->getPointerInstance()->determineSeekPosition();
	}

	/**
	 * Seek to given offset (default) or other possibilities as fseek() gives.
	 *
	 * @param	$offset		Offset to seek to (or used as "base" for other seeks)
	 * @param	$whence		Added to offset (default: only use offset to seek to)
	 * @return	$status		Status of file seek: 0 = success, -1 = failed
	 */
	public function seek (int $offset, $whence = SEEK_SET) {
		// Call pointer instance
		return $this->getPointerInstance()->seek($offset, $whence);
	}

	/**
	 * Reads given amount of bytes from file.
	 *
	 * @param	$bytes	Amount of bytes to read
	 * @return	$data	Data read from file
	 */
	public function read (int $bytes = 0) {
		// Call pointer instance
		return $this->getPointerInstance()->read($bytes);
	}

	/**
	 * Rewinds to the beginning of the file
	 *
	 * @return	$status		Status of this operation
	 */
	public function rewind () {
		// Call pointer instance
		return $this->getPointerInstance()->rewind();
	}

	/**
	 * Analyzes entries in index file. This will count all found (and valid)
	 * entries, mark invalid as damaged and count gaps ("fragmentation"). If
	 * only gaps are found, the file is considered as "virgin" (no entries).
	 *
	 * @return	void
	 * @throws	BadMethodCallException	If this method is called but file is not initialized
	 */
	public function analyzeFile () {
		// Make sure the file is initialized
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		if (!$this->isFileInitialized()) {
			// Bad method call
			throw new BadMethodCallException('Method called but file is not initialized.');
		}

		// Init counters and gaps array
		$this->initCountersGapsArray();

		// Output message (as this may take some time)
		self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('Analyzing file structure ... (this may take some time)'));

		// First rewind to the begining
		$this->rewind();

		// Then try to load all entries
		while ($this->valid()) {
			// Go to next entry
			$this->next();

			// Get current entry
			$current = $this->getCurrentBlock();

			/*
			 * If the block is empty, maybe the whole file is? This could mean
			 * that the file has been pre-allocated.
			 */
			if (empty($current)) {
				// Then skip this part
				continue;
			}

			// Debug message
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('current()=%d', strlen($current)));
		}

		// If the last read block is empty, check gaps
		if (empty($current)) {
			// Output message
			self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('Found a total of %s gaps.', count($this->gaps)));

			// Check gaps, if the whole file is empty.
			if ($this->isFileOnlyGaps()) {
				// Only gaps, so don't continue here.
				return;
			}

			/*
			 * The above call has calculated a total size of all gaps. If the
			 * percentage of gaps passes a "soft" limit and last
			 * defragmentation is to far in the past, or if a "hard" limit has
			 * reached, run defragmentation.
			 */
			if ($this->isDefragmentationNeeded()) {
				// Run "defragmentation"
				$this->doRunDefragmentation();
			}
		}
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Advances to next "block" of bytes
	 *
	 * @return	void
	 * @throws	UnexpectedValueException	If some unexpected value was found
	 */
	public function next () {
		// Is there nothing to read?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		if (!$this->valid()) {
			// Nothing to read
			return;
		}

		// First calculate minimum block length
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: key()=%d', $this->key()));
		$length = $this->getBlockInstance()->calculateMinimumBlockLength();
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: length=%s', $length));

		// Read possibly back-buffered bytes from previous call of next().
		$data = $this->getBackBuffer();

		/*
		 * Read until a entry/block separator has been found. The next read
		 * "block" may not fit, so this loop will continue until the EOB or EOF
		 * has been reached whatever comes first.
		 */
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: data()=%d', strlen($data)));
		while ((!$this->isEndOfFileReached()) && (!self::isBlockSeparatorFound($data))) {
			// Then read the next possible block
			$block = $this->read($length);

			// Is it all empty?
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: block()=%d,length=%s', strlen($block), $length));
			if (strlen(trim($block)) == 0) {
				// Mark this block as empty
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Calling markCurrentBlockAsEmpty(%d) ...', strlen($block)));
				$this->markCurrentBlockAsEmpty(strlen($block));

				// Skip to next block
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CONTINUE!');
				continue;
			}

			// At this block then
			$data .= $block;

			// A debug message
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: data()=%d', strlen($data)));
		}

		// EOF reached?
		if ($this->isEndOfFileReached()) {
			// Set whole data as current read block
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: Calling setCurrentBlock(%d) ...', strlen($data)));
			$this->setCurrentBlock($data);

			// Then abort here silently
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EOF reached - EXIT!');
			return;
		}

		/*
		 * Init back-buffer which is the data that has been found beyond the
		 * separator.
		 */
		$this->initBackBuffer();

		// Separate data
		$dataArray = explode(chr(self::SEPARATOR_ENTRIES), $data);

		// This array must contain two elements
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: dataArray=' . print_r($dataArray, true));
		if (count($dataArray) != 2) {
			// Bad count
			throw new UnexpectedValueException(sprintf('dataArray()=%d is not expected, want 2', count($dataArray)));
		}

		// Left part is the actual block, right one the back-buffer data
		$this->setCurrentBlock($dataArray[0]);
		$this->setBackBuffer($dataArray[1]);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: EXIT!');
	}

	/**
	 * Checks wether the current entry is valid (not at the end of the file).
	 * This method will return true if an emptied (nulled) entry has been found.
	 *
	 * @return	$isValid	Whether the next entry is valid
	 */
	public function valid () {
		// First calculate minimum block length
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-BINARY-FILE: CALLED!');
		$length = $this->getBlockInstance()->calculateMinimumBlockLength();

		// Short be more than zero!
		assert($length > 0);

		// Get current seek position
		$seekPosition = $this->key();

		// Then try to read it
		$data = $this->read($length);

		// If some bytes could be read, all is fine
		$isValid = ((is_string($data)) && (strlen($data) > 0));

		// Get header size
		$headerSize = $this->getHeaderSize();

		// Is the seek position at or beyond the header?
		if ($seekPosition >= $headerSize) {
			// Seek back to old position
			$this->seek($seekPosition);
		} else {
			// Seek directly behind the header
			$this->seek($headerSize);
		}

		// Return result
		return $isValid;
	}

	/**
	 * Gets current seek position ("key").
	 *
	 * @return	$key	Current key in iteration
	 */
	public function key () {
		// Call pointer instance
		return $this->getPointerInstance()->determineSeekPosition();
	}

	/**
	 * Reads the file header
	 *
	 * @return	void
	 */
	public function readFileHeader () {
		// Call block instance
		$this->getBlockInstance()->readFileHeader();
	}

	/**
	 * Flushes the file header
	 *
	 * @return	void
	 */
	public function flushFileHeader () {
		// Call block instance
		$this->getBlockInstance()->flushFileHeader();
	}

	/**
	 * Searches for next suitable gap the given length of data can fit in
	 * including padding bytes.
	 *
	 * @param	$length			Length of raw data
	 * @return	$seekPosition	Found next gap's seek position
	 * @throws	InvalidArgumentException	If the parameter is not valid
	 */
	public function searchNextGap (int $length) {
		// If the file is only gaps, no need to seek
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-BINARY-FILE: length=%d - CALLED!', $length));
		if ($length <= 0) {
			// Throw IAE
			throw new InvalidArgumentException(sprintf('length=%d is not valid', $length));
		} elseif ($this->isFileOnlyGaps()) {
			// The first empty block is the first one right after the header
			return ($this->getHeaderSize() + 1);
		}

		// @TODO Unfinished
		$this->partialStub('length=' . $length);
	}

}
