<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Filesystem\File;

// Import framework stuff
use Org\Mxchange\CoreFramework\Filesystem\CloseableFile;
use Org\Mxchange\CoreFramework\Filesystem\FilePointer;
use Org\Mxchange\CoreFramework\Generic\UnsupportedOperationException;
use Org\Mxchange\CoreFramework\Object\BaseFrameworkSystem;

/**
 * An abstract file class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2020 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
abstract class BaseAbstractFile extends BaseFrameworkSystem implements FilePointer, CloseableFile {
	/**
	 * Counter for total entries
	 */
	private $totalEntries = 0;

	/**
	 * An instance of a file I/O pointer class (not handler)
	 */
	private $pointerInstance = NULL;

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the class
	 * @return	void
	 */
	protected function __construct (string $className) {
		// Call parent constructor
		parent::__construct($className);
	}

	/**
	 * Setter for FilePointer instance
	 *
	 * @param	$pointerInstance	An instance of an FilePointer class
	 * @return	void
	 */
	protected final function setPointerInstance (FilePointer $pointerInstance = NULL) {
		$this->pointerInstance = $pointerInstance;
	}

	/**
	 * Getter for FilePointer instance
	 *
	 * @return	$pointerInstance	An instance of an FilePointer class
	 */
	public final function getPointerInstance () {
		return $this->pointerInstance;
	}

	/**
	 * Unsets pointer instance which triggers a call of __destruct() if the
	 * instance is still there. This is surely not fatal on already "closed"
	 * file pointer instances.
	 *
	 * I don't want to mess around with above setter by giving it a default
	 * value NULL as setter should always explicitly only set (existing) object
	 * instances and NULL is NULL.
	 *
	 * @return	void
	 */
	protected final function unsetPointerInstance () {
		// Simply invoke setter with no parameter
		$this->setPointerInstance();
	}

	/**
	 * Destructor for cleaning purposes, etc
	 *
	 * @return	void
	 */
	public final function __destruct() {
		// Try to close a file
		$this->closeFile();

		// Call the parent destructor
		parent::__destruct();
	}

	/**
	 * "Getter" for abstracted file size
	 *
	 * @return	$fileSize	Size of abstracted file
	 */
	public function getFileSize () {
		// Call pointer instanze
		return $this->getPointerInstance()->getFileSize();
	}

	/**
	 * Getter for total entries
	 *
	 * @return	$totalEntries	Total entries in this file
	 */
	public final function getCounter () {
		// Get it
		return $this->totalEntries;
	}

	/**
	 * Setter for total entries
	 *
	 * @param	$totalEntries	Total entries in this file
	 * @return	void
	 */
	protected final function setCounter (int $counter) {
		// Set it
		$this->totalEntries = $counter;
	}

	/**
	 * Increment counter
	 *
	 * @return	void
	 */
	protected final function incrementCounter () {
		// Get it
		$this->totalEntries++;
	}

	/**
	 * Getter for the file object
	 *
	 * @return	$fileObject		An instance of a SplFileObject
	 */
	public final function getFileObject () {
		return $this->getPointerInstance()->getFileObject();
	}

	/**
	 * Getter for file's name
	 */
	public final function getFilename () {
		// Invole file object's method
		return $this->getFileObject()->getFilename();
	}

	/**
	 * Close a file source and set it's instance to null and the file name
	 * to empty
	 *
	 * @return	void
	 */
	public function closeFile () {
		// Close down pointer instance as well by unsetting it
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-ABSTRACT-FILE: CALLED!');
		$this->unsetPointerInstance();

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-ABSTRACT-FILE: EXIT!');
	}

	/**
	 * Size of this file
	 *
	 * @return	$size	Size (in bytes) of file
	 * @todo	Handle seekStatus
	 */
	public function size () {
		// Call pointer instance
		return $this->getPointerInstance()->size();
	}

	/**
	 * Read data a file pointer
	 *
	 * @return	mixed	The result of fread()
	 * @throws	NullPointerException	If the file pointer instance
	 *									is not set by setFileObject()
	 * @throws	InvalidResourceException	If there is being set
	 */
	public function readFromFile () {
		// Call pointer instance
		return $this->getPointerInstance()->readFromFile();
	}

	/**
	 * Write data to a file pointer
	 *
	 * @param	$dataStream		The data stream we shall write to the file
	 * @return	mixed			Number of writes bytes or false on error
	 * @throws	NullPointerException	If the file pointer instance
	 *									is not set by setFileObject()
	 * @throws	InvalidResourceException	If there is being set
	 *											an invalid file resource
	 */
	public function writeToFile (string $dataStream) {
		// Call pointer instance
		return $this->getPointerInstance()->writeToFile($dataStream);
	}

	/**
	 * Determines whether the EOF has been reached
	 *
	 * @return	$isEndOfFileReached		Whether the EOF has been reached
	 */
	public final function isEndOfFileReached () {
		// Call pointer instance
		return $this->getPointerInstance()->isEndOfFileReached();
	}

}
