<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Index\Stack;

// Import framework stuff
use Org\Mxchange\CoreFramework\Index\BaseIndex;
use Org\Mxchange\CoreFramework\Index\Indexable;
use Org\Mxchange\CoreFramework\Registry\Registerable;
use Org\Mxchange\CoreFramework\Stack\File\StackableFile;
use Org\Mxchange\CoreFramework\Stack\Index\IndexableStack;

// Import SPL stuff
use \InvalidArgumentException;
use \SplFileInfo;

/**
 * A FileStack index class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2020 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class FileStackIndex extends BaseIndex implements IndexableStack, Registerable {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	protected function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates an instance of this Index class and prepares it for usage
	 *
	 * @param	$fileInfoInstance	An instance of a SplFileInfo class
	 * @return	$indexInstance	An instance of this Index class
	 */
	public final static function createFileStackIndex (SplFileInfo $fileInfoInstance) {
		// Get a new instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: fileInfoInstance[%s]=%s - CALLED!', get_class($fileInfoInstance), $fileInfoInstance));
		$indexInstance = new FileStackIndex();

		// Initialize index
		$indexInstance->initIndex($fileInfoInstance);

		// Return the prepared instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: indexInstance=%s - EXIT!', $indexInstance->__toString()));
		return $indexInstance;
	}

	/**
	 * Adds given hash to an index file
	 *
	 * @param	$groupId	Name of stack to add hash for
	 * @param	$data		Hash and gap position to be added to the index
	 * @return	void
	 */
	public function addHashToIndex (string $groupId, array $data) {
		// Raw data been written to the file
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: groupId=%s,hash=%s,gap=%d,length=%d - CALLED!', $groupId, $data[StackableFile::ARRAY_NAME_HASH], $data[StackableFile::ARRAY_NAME_GAP_POSITION], $data[StackableFile::ARRAY_NAME_DATA_LENGTH]));
		$rawData = sprintf('%s%s%s%s%s%s%s',
			$groupId,
			self::SEPARATOR_GROUP_HASH,
			hex2bin($data[StackableFile::ARRAY_NAME_HASH]),
			self::SEPARATOR_HASH_GAP_POSITION,
			$data[StackableFile::ARRAY_NAME_GAP_POSITION],
			self::SEPARATOR_GAP_LENGTH,
			$data[StackableFile::ARRAY_NAME_DATA_LENGTH]
		);

		// Debug message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: groupId=%s,hash=%s,rawData()=%d', $groupId, $data[StackableFile::ARRAY_NAME_HASH], strlen($rawData)));

		// Search for next free gap
		$gapPosition = $this->getIteratorInstance()->searchNextGap(strlen($rawData));

		// Gap position cannot be smaller than header length + 1
		assert($gapPosition > $this->getIteratorInstance()->getHeaderSize());

		// Then write the data at that gap
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: groupId=%s,hash=%s,gapPosition=%s', $groupId, $data[StackableFile::ARRAY_NAME_HASH], $gapPosition));
		$this->getIteratorInstance()->writeData($gapPosition, $rawData);

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: groupId=%s,hash=%s,rawData()=%d - EXIT!', $groupId, $data[StackableFile::ARRAY_NAME_HASH], strlen($rawData)));
	}

	/**
	 * Searches for next suitable gap the given length of data can fit in
	 * including padding bytes.
	 *
	 * @param	$length			Length of raw data
	 * @return	$seekPosition	Found next gap's seek position
	 * @throws	InvalidArgumentException	If the parameter is not valid
	 * @todo	Unfinished work
	 */
	public function searchNextGap (int $length) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('FILE-STACK-INDEX: length=%d - CALLED!', $length));
		if ($length <= 0) {
			// Throw IAE
			throw new InvalidArgumentException(sprintf('length=%d is not valid', $length));
		}
	}

}
